==> 0.4.2&file=in_ffsox-2a-0.4.2-src.tar.gz: in_ffsox-2a-0.4.2/libffsox/ffsox_clone.c <==
/*
 * ffsox_clone.c
 * Copyright (C) 2011-2014 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef NDEBUG /* N.B. assert used with active statements so enable always. */
#undef NDEBUG /* Must undef above assert.h or other that might include it. */
#endif

#include <assert.h>
#include <math.h>
#include "ffsox.h"

#ifndef WIN32
  #define stricmp strcasecmp
  #define strnicmp strncasecmp
#endif

///////////////////////////////////////////////////////////////////////////////
typedef struct ffo_st {
  int skip;
  AVStream *st;
  AVCodec *enc;
  AVCodecContext *ec;
  int index;
  int64_t idts;
} ffo_st_t;

static ffo_st_t *ffo_st_clone(ffo_st_t *ost, ffi_st_t *ist,
    AVFormatContext *oc);
static ffo_st_t *ffo_st_cleanup(ffo_st_t *ost);

///////////////////////////////////////////////////////////////////////////////
typedef struct ffo {
  ffi_t *ffi;
  AVFormatContext *oc;
  ffo_st_t *st;
  ffo_st_t *ast;
} ffo_t;

static ffo_t *ffo_init(ffo_t *ffo, ffi_t *ffi, pb_str_t *fn, char **meta);
static ffo_t *ffo_cleanup(ffo_t *ffo);

static int ffo_loop(ffo_t *ffo);

///////////////////////////////////////////////////////////////////////////////
#if ! defined (AV_CODEC_FLAG_GLOBAL_HEADER) // [
#define AV_CODEC_FLAG_GLOBAL_HEADER CODEC_FLAG_GLOBAL_HEADER
#endif // ]

static ffo_st_t *ffo_st_clone(ffo_st_t *ost, ffi_st_t *ist,
    AVFormatContext *oc)
{
  AVCodecContext *dc=ist->dc;  /* decoding context */
  AVCodecContext *ec;          /* encoding context */
  AVStream *st;
  AVRational sar;

  if (NULL==(ost->st=st=avformat_new_stream(oc,dc->codec)))
    goto error;

  ost->skip=0;
  ost->index=st->index;
  ost->ec=ec=st->codec;

  if (avcodec_copy_context(ec,dc)<0)
    goto error;

#if 1 // {
  // taken from "ffmpeg.c".
  if (AVMEDIA_TYPE_VIDEO==ec->codec_type) {
    if (ist->st->sample_aspect_ratio.num)
      sar=ist->st->sample_aspect_ratio;
    else
      sar=dc->sample_aspect_ratio;

    ost->st->sample_aspect_ratio=ec->sample_aspect_ratio=sar;
    ost->st->avg_frame_rate=ist->st->avg_frame_rate;
  }
#endif // }

  ec->codec_tag=0;

  if (oc->oformat->flags&AVFMT_GLOBALHEADER)
    ec->flags|=AV_CODEC_FLAG_GLOBAL_HEADER;

  return ost;
error:
  return NULL;
}

static ffo_st_t *ffo_st_cleanup(ffo_st_t *ost)
{
  return ost;
}

///////////////////////////////////////////////////////////////////////////////
static void ffo_clone_dict_filtered(AVDictionary **od, AVDictionary *id,
    char **meta)
{
  AVDictionaryEntry *mtag=NULL;
  char **m;

  while (NULL!=(mtag=av_dict_get(id,"",mtag,AV_DICT_IGNORE_SUFFIX))) {
    for (m=meta;m&&*m;++m)
      if (!stricmp(*m,mtag->key))
        goto next_mtag;

    av_dict_set(od,mtag->key,mtag->value,0);
  next_mtag:
    continue;
  }
}

static void ffo_add_dict(AVDictionary **od, char **meta)
{
  char **m;

  for (m=meta;m&&*m;++m) {
    char *val=*m+strlen(*m)+1;

    if (*val)
      av_dict_set(od,*m,val,0);
  }
}

static ffo_t *ffo_init(ffo_t *ffo, ffi_t *ffi, pb_str_t *fn, char **meta)
{
  AVOutputFormat *oformat=NULL;
  AVFormatContext *oc=NULL;
  int i;

  memset(ffo,0,sizeof *ffo);
  ffo->ffi=ffi;

  if (avformat_alloc_output_context2(&oc,NULL,NULL,PS_UTF8(fn))<0)
    goto error;

  ffo->oc=oc;
  oformat=oc->oformat;

  if (NULL==(ffo->st=calloc(ffi->nb_streams, sizeof *ffo->st)))
    goto error;

  for (i=0;i<ffi->nb_streams;++i) {
    ffi_st_t *ist=ffi->st+i;
    ffo_st_t *ost=ffo->st+i;
    AVCodecContext *dc=ist->dc;

    if (NULL==dc
        ||(AVMEDIA_TYPE_VIDEO==dc->codec_type
        &&AV_CODEC_ID_NONE==oformat->video_codec)) {

      ost->skip=1;
      continue;
    }
    else if (NULL==ffo_st_clone(ost,ist,oc))
      goto error;

    if (ffi->ast==ist)
      ffo->ast=ost;
  }

  ffo_clone_dict_filtered(&oc->metadata,ffi->ic->metadata,meta);

#if defined (FFO_CLONE_STREAM_META) // {
  for (i=0;i<ffi->nb_streams;++i) {
    ffi_st_t *ist=ffi->st+i;
    ffo_st_t *ost=ffo->st+i;

    ffo_clone_dict_filtered(&oc->metadata,ist->st->metadata,meta);
  }
#endif // }

  ffo_add_dict(&oc->metadata,meta);

  if (0!=(oformat->flags&AVFMT_NOFILE)
      ||avio_open(&oc->pb,PS_UTF8(fn),AVIO_FLAG_WRITE)<0)
    goto error;

  return ffo;
error:
#if defined (WIN32)
  _ftprintf(stderr,_T("Error opening \"%s\".\n"),PS_TCS(fn));
#else
  fprintf(stderr,"Error opening \"%s\".\n",PS_UTF8(fn));
#endif
  ffo_cleanup(ffo);

  return NULL;
}

static ffo_t *ffo_cleanup(ffo_t *ffo)
{
  AVFormatContext *oc=ffo->oc;
  int i;

  if (NULL!=oc) {
    if (NULL!=oc->pb&&!(oc->flags&AVFMT_NOFILE)) {
      avio_close(oc->pb);
      oc->pb=NULL;
    }

    avformat_free_context(oc);
    ffo->oc=NULL;
  }

  if (NULL!=ffo->st) {
    free(ffo->st);
    ffo->st=NULL;
  }

  return ffo;
}

static int ffo_loop(ffo_t *ffo)
{
  ffi_t *ffi=ffo->ffi;
  AVFormatContext *ic=ffi->ic;
  AVFormatContext *oc=ffo->oc;
  AVPacket pkt;
  ffi_st_t *ist;
  ffo_st_t *ost;
  int count,ret;

  if (0!=avformat_write_header(oc,NULL))
    return -1;

  count=0;

  while (0==av_read_frame(ic,&pkt)) {
    ret=-1;

    if (ffi->nb_streams<=pkt.stream_index
        ||NULL==(ist=ffi->st+pkt.stream_index)->st
        ||NULL==(ost=ffo->st+pkt.stream_index)->st) {

      goto pkt;
    }

#if 0 // { libav
    pkt.pts=av_rescale_q_rnd(pkt.pts,ist->st->time_base,ost->st->time_base,
        AV_ROUND_NEAR_INF);
    pkt.dts=av_rescale_q_rnd(pkt.dts,ist->st->time_base,ost->st->time_base,
        AV_ROUND_NEAR_INF);
#else // } { ffmpeg
    pkt.pts=av_rescale_q_rnd(pkt.pts,ist->st->time_base,ost->st->time_base,
        AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
    pkt.dts=av_rescale_q_rnd(pkt.dts,ist->st->time_base,ost->st->time_base,
        AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
#endif // }
    pkt.duration=av_rescale_q(pkt.duration,ist->st->time_base,
        ost->st->time_base);
    pkt.pos=-1;
    ret=av_interleaved_write_frame(oc,&pkt);
  pkt:
    av_free_packet(&pkt);

    if (0!=ret)
      return ret;

    ++count;
  }

  av_write_trailer(oc);
  avio_flush(oc->pb);

  return 0==count?-1:0;
}

///////////////////////////////////////////////////////////////////////////////
#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
int ffsox_clone(pb_str_t *fn1, pb_str_t *fn2, int mode, char **meta,
    double *scale)
#else // ] [
int ffsox_clone(pb_str_t *fn1, pb_str_t *fn2, int mode, char **meta)
#endif // ]
{
  int code=-1;
  ffi_t ffi;
  ffo_t ffo;

  memset(&ffi,0,sizeof ffi);
  memset(&ffo,0,sizeof ffo);

#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
  if (NULL==ffi_init(&ffi,fn1,mode|FFI_STREAM_META|FFI_OPEN,scale))
    goto cleanup;
#else // ] [
  if (NULL==ffi_init(&ffi,fn1,mode|FFI_STREAM_META|FFI_OPEN))
    goto cleanup;
#endif // ]

  if (NULL==ffo_init(&ffo,&ffi,fn2,meta))
    goto cleanup;

  code=ffo_loop(&ffo);
cleanup:
  ffo_cleanup(&ffo);
  ffi_cleanup(&ffi);

  if (0!=code) {
#if defined (WIN32)
    _tunlink(fn2->tcs);
#else
    unlink(fn2->utf8);
#endif
  }

  return code;
}

==> 0.6.0&file=in_ffsox-2a-0.6.0-src.tar.gz: in_ffsox-2a-0.6.0/libffsox/ffsox_clone.c <==
/*
 * ffsox_clone.c
 * Copyright (C) 2011-2014 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef NDEBUG /* N.B. assert used with active statements so enable always. */
#undef NDEBUG /* Must undef above assert.h or other that might include it. */
#endif

#include <assert.h>
#include <math.h>
#include "ffsox.h"

#ifndef WIN32
  #define stricmp strcasecmp
  #define strnicmp strncasecmp
#endif

///////////////////////////////////////////////////////////////////////////////
typedef struct ffo_st {
  int skip;
  AVStream *st;
  AVCodec *enc;
  AVCodecContext *ec;
  int index;
  int64_t idts;
} ffo_st_t;

static ffo_st_t *ffo_st_clone(ffo_st_t *ost, ffi_st_t *ist,
    AVFormatContext *oc);
static ffo_st_t *ffo_st_cleanup(ffo_st_t *ost);

///////////////////////////////////////////////////////////////////////////////
typedef struct ffo {
  ffi_t *ffi;
  AVFormatContext *oc;
  int n;
  ffo_st_t *ost;
  ffo_st_t *ast;
} ffo_t;

static ffo_t *ffo_init(ffo_t *ffo, ffi_t *ffi, pb_str_t *fn, char **meta);
static ffo_t *ffo_cleanup(ffo_t *ffo);

static int ffo_loop(ffo_t *ffo);

///////////////////////////////////////////////////////////////////////////////
#if ! defined (AV_CODEC_FLAG_GLOBAL_HEADER) // [
#define AV_CODEC_FLAG_GLOBAL_HEADER CODEC_FLAG_GLOBAL_HEADER
#endif // ]

static ffo_st_t *ffo_st_clone(ffo_st_t *ost, ffi_st_t *ist,
    AVFormatContext *oc)
{
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  AVCodecParameters *codecpar=ist->st->codecpar;
#else // ] [
  AVCodecContext *codecpar=ist->st->codec;
#endif // ]
  AVCodecContext *ec;
  AVStream *st;
  AVRational sar;

  // find decoder for the stream.
  if (NULL==(ost->enc=ffi_find_decoder(codecpar->codec_id)))
    goto error;

  if (NULL==(ost->st=st=avformat_new_stream(oc,ost->enc)))
    goto error;

#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  // Allocate a codec context for the decoder.
  if (!(ost->ec=ec=avcodec_alloc_context3(ost->enc)))
    goto error;

  // Copy codec parameters from input stream to output codec context.
  if (avcodec_parameters_to_context(ost->ec,codecpar)<0)
    goto error;
#else // ] [
  ost->ec=ec=ost->st->codec;

  if (avcodec_copy_context(ost->ec,codecpar)<0)
    goto error;
#endif // ]

  ost->skip=0;
  ost->index=st->index;
#if 0 // [
  ost->ec=ec=st->codec;
#endif // ]

#if 1 // {
  // taken from "ffmpeg.c".
  if (AVMEDIA_TYPE_VIDEO==ec->codec_type) {
    if (ist->st->sample_aspect_ratio.num)
      sar=ist->st->sample_aspect_ratio;
    else
      sar=codecpar->sample_aspect_ratio;

    ost->st->sample_aspect_ratio=ec->sample_aspect_ratio=sar;
    ost->st->avg_frame_rate=ist->st->avg_frame_rate;
  }
#endif // }

  ec->codec_tag=0;

  if (oc->oformat->flags&AVFMT_GLOBALHEADER)
    ec->flags|=AV_CODEC_FLAG_GLOBAL_HEADER;

  return ost;
error:
  return NULL;
}

static ffo_st_t *ffo_st_cleanup(ffo_st_t *ost)
{
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  if (!ost->skip)
    avcodec_free_context(&ost->ec);
#endif // ]

  return ost;
}

///////////////////////////////////////////////////////////////////////////////
static void ffo_clone_dict_filtered(AVDictionary **od, AVDictionary *id,
    char **meta)
{
  AVDictionaryEntry *mtag=NULL;
  char **m;

  while (NULL!=(mtag=av_dict_get(id,"",mtag,AV_DICT_IGNORE_SUFFIX))) {
    for (m=meta;m&&*m;++m)
      if (!stricmp(*m,mtag->key))
        goto next_mtag;

    av_dict_set(od,mtag->key,mtag->value,0);
  next_mtag:
    continue;
  }
}

static void ffo_add_dict(AVDictionary **od, char **meta)
{
  char **m;

  for (m=meta;m&&*m;++m) {
    char *val=*m+strlen(*m)+1;

    if (*val)
      av_dict_set(od,*m,val,0);
  }
}

static ffo_t *ffo_init(ffo_t *ffo, ffi_t *ffi, pb_str_t *fn, char **meta)
{
  AVOutputFormat *oformat=NULL;
  AVFormatContext *oc=NULL;

  memset(ffo,0,sizeof *ffo);
  ffo->ffi=ffi;

  if (avformat_alloc_output_context2(&oc,NULL,NULL,PS_UTF8(fn))<0)
    goto error;

  ffo->oc=oc;
  oformat=oc->oformat;

  if (NULL==(ffo->ost=calloc(ffi->nb_streams, sizeof *ffo->ost)))
    goto error;

  for (ffo->n=0;ffo->n<ffi->nb_streams;++ffo->n) {
    ffi_st_t *ist=ffi->ist+ffo->n;
    ffo_st_t *ost=ffo->ost+ffo->n;
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
    AVCodecParameters *codecpar=ist->st->codecpar;
#else // ] [
    AVCodecContext *codecpar=ist->st->codec;
#endif // ]

    if (NULL==codecpar
        ||(AVMEDIA_TYPE_VIDEO==codecpar->codec_type
        &&AV_CODEC_ID_NONE==oformat->video_codec)) {

      ost->skip=1;
      continue;
    }
    else if (NULL==ffo_st_clone(ost,ist,oc))
      goto error;

    if (ffi->ast==ist)
      ffo->ast=ost;
  }

  ffo_clone_dict_filtered(&oc->metadata,ffi->ic->metadata,meta);

#if defined (FFO_CLONE_STREAM_META) // {
  for (i=0;i<ffi->nb_streams;++i) {
    ffi_st_t *ist=ffi->ist+i;
    ffo_st_t *ost=ffo->st+i;

    ffo_clone_dict_filtered(&oc->metadata,ist->st->metadata,meta);
  }
#endif // }

  ffo_add_dict(&oc->metadata,meta);

  if (0!=(oformat->flags&AVFMT_NOFILE)
      ||avio_open(&oc->pb,PS_UTF8(fn),AVIO_FLAG_WRITE)<0)
    goto error;

  return ffo;
error:
#if defined (WIN32)
  _ftprintf(stderr,_T("Error opening \"%s\".\n"),PS_TCS(fn));
#else
  fprintf(stderr,"Error opening \"%s\".\n",PS_UTF8(fn));
#endif
  ffo_cleanup(ffo);

  return NULL;
}

static ffo_t *ffo_cleanup(ffo_t *ffo)
{
  AVFormatContext *oc=ffo->oc;
  int i;

  if (NULL!=oc) {
    for (i=0;i<ffo->n;++i)
      ffo_st_cleanup(ffo->ost+i);

    if (NULL!=oc->pb&&!(oc->flags&AVFMT_NOFILE)) {
      avio_close(oc->pb);
      oc->pb=NULL;
    }

    avformat_free_context(oc);
    ffo->oc=NULL;
  }

  if (NULL!=ffo->ost) {
    free(ffo->ost);
    ffo->ost=NULL;
  }

  return ffo;
}

static int ffo_loop(ffo_t *ffo)
{
  ffi_t *ffi=ffo->ffi;
  AVFormatContext *ic=ffi->ic;
  AVFormatContext *oc=ffo->oc;
  AVPacket pkt;
  ffi_st_t *ist;
  ffo_st_t *ost;
  int count,ret;

  if (0!=avformat_write_header(oc,NULL))
    return -1;

  count=0;

  while (0==av_read_frame(ic,&pkt)) {
    ret=-1;

    if (ffi->nb_streams<=pkt.stream_index
        ||NULL==(ist=ffi->ist+pkt.stream_index)->st
        ||NULL==(ost=ffo->ost+pkt.stream_index)->st) {

      goto pkt;
    }

#if 0 // { libav
    pkt.pts=av_rescale_q_rnd(pkt.pts,ist->st->time_base,ost->st->time_base,
        AV_ROUND_NEAR_INF);
    pkt.dts=av_rescale_q_rnd(pkt.dts,ist->st->time_base,ost->st->time_base,
        AV_ROUND_NEAR_INF);
#else // } { ffmpeg
    pkt.pts=av_rescale_q_rnd(pkt.pts,ist->st->time_base,ost->st->time_base,
        AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
    pkt.dts=av_rescale_q_rnd(pkt.dts,ist->st->time_base,ost->st->time_base,
        AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
#endif // }
    pkt.duration=av_rescale_q(pkt.duration,ist->st->time_base,
        ost->st->time_base);
    pkt.pos=-1;
    ret=av_interleaved_write_frame(oc,&pkt);
  pkt:
#if defined (FFSOX_DEPRECATED_FREE_PACKET) // [
    av_packet_unref(&pkt);
#else // ] [
    av_free_packet(&pkt);
#endif // ]

    if (0!=ret)
      return ret;

    ++count;
  }

  av_write_trailer(oc);
  avio_flush(oc->pb);

  return 0==count?-1:0;
}

///////////////////////////////////////////////////////////////////////////////
#if defined (FFSOX_FFI_DECODE) // [
int ffsox_clone(pb_str_t *fn1, pb_str_t *fn2, int mode, char **meta,
    ffi_st_cbs_t *cbs, void *data)
#elif defined (FFSOX_AUDIO_DRC) // [
int ffsox_clone(pb_str_t *fn1, pb_str_t *fn2, int mode, char **meta,
    double *scale)
#else // ] [
int ffsox_clone(pb_str_t *fn1, pb_str_t *fn2, int mode, char **meta)
#endif // ]
{
  int code=-1;
  ffi_t ffi;
  ffo_t ffo;

  memset(&ffi,0,sizeof ffi);
  memset(&ffo,0,sizeof ffo);

#if defined (FFSOX_MAP_MEMORY) // [
#if defined (FFSOX_FFI_DECODE) // [
  if (NULL==ffi_init(&ffi,fn1,0,mode|FFI_STREAM_META|FFI_OPEN,cbs,data))
    goto cleanup;
#elif defined (FFSOX_AUDIO_DRC) // [
  if (NULL==ffi_init(&ffi,fn1,0,mode|FFI_STREAM_META|FFI_OPEN,scale))
    goto cleanup;
#else // ] [
  if (NULL==ffi_init(&ffi,fn1,0,mode|FFI_STREAM_META|FFI_OPEN))
    goto cleanup;
#endif // ]
#else // ] [
#if defined (FFSOX_FFI_DECODE) // [
  if (NULL==ffi_init(&ffi,fn1,mode|FFI_STREAM_META|FFI_OPEN,cbs,data))
    goto cleanup;
#elif defined (FFSOX_AUDIO_DRC) // [
  if (NULL==ffi_init(&ffi,fn1,mode|FFI_STREAM_META|FFI_OPEN,scale))
    goto cleanup;
#else // ] [
  if (NULL==ffi_init(&ffi,fn1,mode|FFI_STREAM_META|FFI_OPEN))
    goto cleanup;
#endif // ]
#endif // ]

  if (NULL==ffo_init(&ffo,&ffi,fn2,meta))
    goto cleanup;

  code=ffo_loop(&ffo);
cleanup:
  ffo_cleanup(&ffo);
  ffi_cleanup(&ffi);

  if (0!=code) {
#if defined (WIN32)
    _tunlink(fn2->tcs);
#else
    unlink(fn2->utf8);
#endif
  }

  return code;
}
